@@ -1,7 +1,7 @@
-/* $Id: milter-regex.c,v 1.49 2005/03/06 11:42:53 dhartmei Exp $ */
+/* $Id: milter-regex.c,v 1.3 2007/08/03 22:11:48 dhartmei Exp $ */
 
 /*
- * Copyright (c) 2003-2004 Daniel Hartmeier
+ * Copyright (c) 2003-2006 Daniel Hartmeier
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,7 +30,7 @@
  *
  */
 
-static const char rcsid[] = "$Id: milter-regex.c,v 1.49 2005/03/06 11:42:53 dhartmei Exp $";
+static const char rcsid[] = "$Id: milter-regex.c,v 1.3 2007/08/03 22:11:48 dhartmei Exp $";
 
 #include <sys/types.h>
 #include <sys/socket.h>
@@ -39,6 +39,7 @@ static const char rcsid[] = "$Id: milter-regex.c,v 1.49 2005/03/06 11:42:53 dhar
 #include <arpa/inet.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <pthread.h>
 #include <pwd.h>
 #include <stdarg.h>
 #include <stdio.h>
@@ -53,10 +54,13 @@ static const char rcsid[] = "$Id: milter-regex.c,v 1.49 2005/03/06 11:42:53 dhar
 
 #include "eval.h"
 
-extern int parse_ruleset(const char *, struct ruleset **, char *, size_t);
+extern void	 die(const char *);
+extern int	 parse_ruleset(const char *, struct ruleset **, char *,
+		    size_t);
 
 static const char	*rule_file_name = "/etc/milter-regex.conf";
 static int		 debug = 0;
+static pthread_mutex_t	 mutex;
 
 struct context {
 	struct ruleset	*rs;
@@ -65,6 +69,7 @@ struct context {
 	unsigned	 pos;		/* write position within buf */
 	char		 host_name[128];
 	char		 host_addr[64];
+	char		 helo[128];
 	char		 hdr_from[128];
 	char		 hdr_to[128];
 	char		 hdr_subject[128];
@@ -94,6 +99,38 @@ static void		 msg(int, struct context *, const char *, ...);
 #define XCODE_TEMPFAIL	"4.7.1"
 #define	MAXRS		16
 
+/* Define what sendmail macros should be queried in what context (phase)
+ * with smfi_getsymval(). Whether sendmail actually provides specific
+ * values depends on configuration of confMILTER_MACROS_*
+ */
+struct {
+	const char *phase;
+	const char *name;
+} macro[] = {
+	{ "connect", "{daemon_name}" },
+	{ "connect", "{if_name}" },
+	{ "connect", "{if_addr}" },
+	{ "connect", "j" },
+	{ "connect", "_" },
+	{ "helo", "{tls_version}" },
+	{ "helo", "{cipher}" },
+	{ "helo", "{cipher_bits}" },
+	{ "helo", "{cert_subject}" },
+	{ "helo", "{cert_issuer}" },
+	{ "envfrom", "i" },
+	{ "envfrom", "{auth_type}" },
+	{ "envfrom", "{auth_authen}" },
+	{ "envfrom", "{auth_ssf}" },
+	{ "envfrom", "{auth_author}" },
+	{ "envfrom", "{mail_mailer}" },
+	{ "envfrom", "{mail_host}" },
+	{ "envfrom", "{mail_addr}" },
+	{ "envrcpt", "{rcpt_mailer}" },
+	{ "envrcpt", "{rcpt_host}" },
+	{ "envrcpt", "{rcpt_addr}" },
+	{ NULL, NULL }
+};
+
 #if __linux__ || __sun__
 #define	ST_MTIME st_mtime
 extern size_t	 strlcpy(char *, const char *, size_t);
@@ -101,6 +138,20 @@ extern size_t	 strlcpy(char *, const char *, size_t);
 #define	ST_MTIME st_mtimespec
 #endif
 
+static void
+mutex_lock(void)
+{
+	if (pthread_mutex_lock(&mutex))
+		die("pthread_mutex_lock");
+}
+
+static void
+mutex_unlock(void)
+{
+	if (pthread_mutex_unlock(&mutex))
+		die("pthread_mutex_unlock");
+}
+
 #ifdef __sun__
 int
 daemon(int nochdir, int noclose)
@@ -142,15 +193,15 @@ setreply(SMFICTX *ctx, struct context *context, const struct action *action)
 
 	switch (action->type) {
 	case ACTION_REJECT:
-		msg(LOG_NOTICE, context, "REJECT: %s, From: %s, To: %s, "
-		    "Subject: %s", action->msg, context->hdr_from,
-		    context->hdr_to, context->hdr_subject);
+		msg(LOG_NOTICE, context, "REJECT: %s, Helo: %s, From: %s, "
+		    "To: %s, Subject: %s", action->msg, context->helo,
+		    context->hdr_from, context->hdr_to, context->hdr_subject);
 		result = SMFIS_REJECT;
 		break;
 	case ACTION_TEMPFAIL:
-		msg(LOG_NOTICE, context, "TEMPFAIL: %s, From: %s, To: %s, "
-		    "Subject: %s", action->msg, context->hdr_from,
-		    context->hdr_to, context->hdr_subject);
+		msg(LOG_NOTICE, context, "TEMPFAIL: %s, Helo: %s, From: %s, "
+		    "To: %s, Subject: %s", action->msg, context->helo,
+		    context->hdr_from, context->hdr_to, context->hdr_subject);
 		result = SMFIS_TEMPFAIL;
 		break;
 	case ACTION_QUARANTINE:
@@ -159,15 +210,15 @@ setreply(SMFICTX *ctx, struct context *context, const struct action *action)
 		context->quarantine = strdup(action->msg);
 		break;
 	case ACTION_DISCARD:
-		msg(LOG_NOTICE, context, "DISCARD, From: %s, To: %s, "
-		    "Subject: %s", context->hdr_from, context->hdr_to,
-		    context->hdr_subject);
+		msg(LOG_NOTICE, context, "DISCARD, Helo: %s, From: %s, "
+		    "To: %s, Subject: %s", context->helo, context->hdr_from,
+		    context->hdr_to, context->hdr_subject);
 		result = SMFIS_DISCARD;
 		break;
 	case ACTION_ACCEPT:
-		msg(LOG_INFO, context, "ACCEPT, From: %s, To: %s, "
-		    "Subject: %s", context->hdr_from, context->hdr_to,
-		    context->hdr_subject);
+		msg(LOG_INFO, context, "ACCEPT, Helo: %s, From: %s, "
+		    "To: %s, Subject: %s", context->helo, context->hdr_from,
+		    context->hdr_to, context->hdr_subject);
 		result = SMFIS_ACCEPT;
 		break;
 	}
@@ -192,6 +243,7 @@ get_ruleset(void)
 	time_t t = time(NULL);
 	int load = 0;
 
+	mutex_lock();
 	if (!last_check)
 		memset(&sbo, 0, sizeof(sbo));
 	if (t - last_check >= 10) {
@@ -236,13 +288,35 @@ get_ruleset(void)
 			cur = i;
 		}
 	}
+	mutex_unlock();
 	return (rs[cur]);
 }
 
+static struct action *
+check_macros(SMFICTX *ctx, struct context *context, const char *phase)
+{
+	struct action *action;
+	int i;
+	const char *v;
+
+	for (i = 0; macro[i].phase != NULL; ++i) {
+		if (strcmp(macro[i].phase, phase))
+			continue;
+		if ((v = smfi_getsymval(ctx, (char *)macro[i].name)) == NULL)
+			v = "";
+		msg(LOG_DEBUG, context, "macro %s = %s", macro[i].name, v);
+		if ((action = eval_cond(context->rs, context->res, COND_MACRO,
+		    macro[i].name, v)) != NULL)
+			return (action);
+	}
+	return (NULL);
+}
+
 static sfsistat
 cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
 {
 	struct context *context;
+	struct action *action;
 
 	context = calloc(1, sizeof(*context));
 	if (context == NULL) {
@@ -271,30 +345,36 @@ cb_connect(SMFICTX *ctx, char *name, _SOCK_ADDR *sa)
 
 	strlcpy(context->host_name, name, sizeof(context->host_name));
 	strlcpy(context->host_addr, "unknown", sizeof(context->host_addr));
-	switch (sa->sa_family) {
-	case AF_INET: {
-		struct sockaddr_in *sin = (struct sockaddr_in *)sa;
+	if (sa) {
+		switch (sa->sa_family) {
+		case AF_INET: {
+			struct sockaddr_in *sin = (struct sockaddr_in *)sa;
 
-		if (inet_ntop(AF_INET, &sin->sin_addr.s_addr,
-		    context->host_addr, sizeof(context->host_addr)) ==
-		    NULL)
-			msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
-			    strerror(errno));
-		break;
-	}
-	case AF_INET6: {
-		struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
+			if (inet_ntop(AF_INET, &sin->sin_addr.s_addr,
+			    context->host_addr, sizeof(context->host_addr)) ==
+			    NULL)
+				msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
+				    strerror(errno));
+			break;
+		}
+		case AF_INET6: {
+			struct sockaddr_in6 *sin6 = (struct sockaddr_in6 *)sa;
 
-		if (inet_ntop(AF_INET6, &sin6->sin6_addr,
-		    context->host_addr, sizeof(context->host_addr)) ==
-		    NULL)
-			msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
-			    strerror(errno));
-		break;
-	}
+			if (inet_ntop(AF_INET6, &sin6->sin6_addr,
+			    context->host_addr, sizeof(context->host_addr)) ==
+			    NULL)
+				msg(LOG_ERR, NULL, "cb_connect: inet_ntop: %s",
+				    strerror(errno));
+			break;
+		}
+		}
 	}
 	msg(LOG_DEBUG, context, "cb_connect('%s', '%s')",
 	    context->host_name, context->host_addr);
+	if ((action = check_macros(ctx, context, "connect")) != NULL) {
+		/* can't really do this, delay */
+		/*return (setreply(ctx, context, action)); */
+	}
 	return (SMFIS_CONTINUE);
 }
 
@@ -308,6 +388,8 @@ cb_helo(SMFICTX *ctx, char *arg)
 		msg(LOG_ERR, NULL, "cb_helo: smfi_getpriv");
 		return (SMFIS_ACCEPT);
 	}
+	strlcpy(context->helo, arg, sizeof(context->helo));
+	msg(LOG_DEBUG, context, "cb_helo('%s')", arg);
 	/* multiple HELO imply RSET in sendmail */
 	/* evaluate connect arguments here, because we can't call */
 	/* setreply from cb_connect */
@@ -315,16 +397,18 @@ cb_helo(SMFICTX *ctx, char *arg)
 	if ((action = eval_cond(context->rs, context->res, COND_CONNECT,
 	    context->host_name, context->host_addr)) != NULL)
 		return (setreply(ctx, context, action));
-	if ((action = eval_end(context->rs, context->res, COND_CONNECT)) !=
+	if ((action = eval_end(context->rs, context->res, COND_CONNECT,
+	    COND_MACRO)) !=
 	    NULL)
 		return (setreply(ctx, context, action));
-	msg(LOG_DEBUG, context, "cb_helo('%s')", arg);
+	if ((action = check_macros(ctx, context, "helo")) != NULL)
+		return (setreply(ctx, context, action));
 	eval_clear(context->rs, context->res, COND_HELO);
 	if ((action = eval_cond(context->rs, context->res, COND_HELO,
 	    arg, NULL)) != NULL)
 		return (setreply(ctx, context, action));
-	if ((action = eval_end(context->rs, context->res, COND_HELO)) !=
-	    NULL)
+	if ((action = eval_end(context->rs, context->res, COND_HELO,
+	    COND_MACRO)) != NULL)
 		return (setreply(ctx, context, action));
 	return (SMFIS_CONTINUE);
 }
@@ -347,8 +431,10 @@ cb_envfrom(SMFICTX *ctx, char **args)
 		    *args, NULL)) != NULL)
 			return (setreply(ctx, context, action));
 	}
-	if ((action = eval_end(context->rs, context->res, COND_ENVFROM)) !=
-	    NULL)
+	if ((action = eval_end(context->rs, context->res, COND_ENVFROM,
+	    COND_MACRO)) != NULL)
+		return (setreply(ctx, context, action));
+	if ((action = check_macros(ctx, context, "envfrom")) != NULL)
 		return (setreply(ctx, context, action));
 	return (SMFIS_CONTINUE);
 }
@@ -371,8 +457,10 @@ cb_envrcpt(SMFICTX *ctx, char **args)
 		    *args, NULL)) != NULL)
 			return (setreply(ctx, context, action));
 	}
-	if ((action = eval_end(context->rs, context->res, COND_ENVRCPT)) !=
-	    NULL)
+	if ((action = eval_end(context->rs, context->res, COND_ENVRCPT,
+	    COND_MACRO)) != NULL)
+		return (setreply(ctx, context, action));
+	if ((action = check_macros(ctx, context, "envrcpt")) != NULL)
 		return (setreply(ctx, context, action));
 	return (SMFIS_CONTINUE);
 }
@@ -388,6 +476,9 @@ cb_header(SMFICTX *ctx, char *name, char *value)
 		return (SMFIS_ACCEPT);
 	}
 	msg(LOG_DEBUG, context, "cb_header('%s', '%s')", name, value);
+	if ((action = eval_end(context->rs, context->res, COND_MACRO,
+	    COND_HEADER)) != NULL)
+		return (setreply(ctx, context, action));
 	if (!strcasecmp(name, "From"))
 		strlcpy(context->hdr_from, value, sizeof(context->hdr_from));
 	else if (!strcasecmp(name, "To"))
@@ -414,8 +505,8 @@ cb_eoh(SMFICTX *ctx)
 	msg(LOG_DEBUG, context, "cb_eoh()");
 	memset(context->buf, 0, sizeof(context->buf));
 	context->pos = 0;
-	if ((action = eval_end(context->rs, context->res, COND_HEADER)) !=
-	    NULL)
+	if ((action = eval_end(context->rs, context->res, COND_HEADER,
+	    COND_BODY)) != NULL)
 		return (setreply(ctx, context, action));
 	return (SMFIS_CONTINUE);
 }
@@ -463,17 +554,17 @@ cb_eom(SMFICTX *ctx)
 		return (SMFIS_ACCEPT);
 	}
 	msg(LOG_DEBUG, context, "cb_eom()");
-	if ((action = eval_end(context->rs, context->res, COND_BODY)) !=
-	    NULL)
+	if ((action = eval_end(context->rs, context->res, COND_BODY,
+	    COND_MAX)) != NULL)
 		result = setreply(ctx, context, action);
 	else
-		msg(LOG_DEBUG, context, "ACCEPT, From: %s, To: %s, "
-		    "Subject: %s", context->hdr_from, context->hdr_to,
-		    context->hdr_subject);
-	if (context->quarantine != NULL) {
-		msg(LOG_NOTICE, context, "QUARANTINE: %s, From: %s, To: %s, "
-		    "Subject: %s", action->msg, context->hdr_from,
+		msg(LOG_DEBUG, context, "ACCEPT, Helo: %s, From: %s, To: %s, "
+		    "Subject: %s", context->helo, context->hdr_from,
 		    context->hdr_to, context->hdr_subject);
+	if (context->quarantine != NULL) {
+		msg(LOG_NOTICE, context, "QUARANTINE: %s, Helo: %s, From: %s, "
+		    "To: %s, Subject: %s", action->msg, context->helo,
+		    context->hdr_from, context->hdr_to, context->hdr_subject);
 		if (smfi_quarantine(ctx, context->quarantine) != MI_SUCCESS)
 			msg(LOG_ERR, context, "cb_eom: smfi_quarantine");
 	}
@@ -618,6 +709,11 @@ main(int argc, char **argv)
 		}
 	}
 
+	if (pthread_mutex_init(&mutex, 0)) {
+		fprintf(stderr, "pthread_mutex_init\n");
+		goto done;
+	}
+
 	if (smfi_setconn((char *)oconn) != MI_SUCCESS) {
 		fprintf(stderr, "smfi_setconn: %s: failed\n", oconn);
 		goto done;
